'''
All shared bits of code that are used throughout much of the UIValidation package
that don't have any better place to be. This encompasses all shared utility code
such as the utility classes, enums, and exceptions used by this application.
'''
